/*
 *	Boot a qnx image built with qnx tools
*/

#include <common.h>
#include <common.h>
#include <watchdog.h>
#include <command.h>
#include <cmd_boot.h>
#include <image.h>
#include <malloc.h>
#include <zlib.h>
#include <asm/byteorder.h>
#if (CONFIG_COMMANDS & CFG_CMD_DATE)
#include <rtc.h>
#endif

/*
 * QNX image header definition
 */
struct startup_header {
	unsigned long	signature;
	unsigned short	version;
	unsigned char	flags1;
	unsigned char	flags2;
	unsigned short	header_size;
	unsigned short	machine;
	unsigned long	startup_vaddr;
										/*   to after IPL is done*/
	unsigned long	paddr_bias;
										/*   to get a value to put into a*/
										/*   pointer and indirected through*/
	unsigned long	image_paddr;
	unsigned long	ram_paddr;
										/*   image to (startup_size bytes copied)*/
	unsigned long	ram_size;
										/*   program and executables contained*/
										/*   in the file system*/
	unsigned long	startup_size;
	unsigned long	stored_size;
	unsigned long	imagefs_paddr;
	unsigned long	imagefs_size;
	unsigned short	preboot_size;
	unsigned short	zero0;
	unsigned long	zero[3];
	unsigned long	info[48];
};

#define STARTUP_HDR_SIGNATURE			0x00ff7eeb
#define STARTUP_HDR_VERSION				1
#define GOLDEN_IMAGE                    0xfff00000

void memory_fill_words(unsigned long dst, int pattern, int size)
{
	int i;
	unsigned int *d = (unsigned int *)dst;

	for ( i=0; i<size/4; i++ ) {
		d[i] = pattern;
	}
}

void verify_fill_words(unsigned long src, int pattern, int size)
{
	int i;
	unsigned int *s = (unsigned int *)src;
	unsigned int kk;

	for ( i=0; i<size/4; i++ ) {
		if ( (kk = s[i]) != pattern)
			printf("verify_fill_words: Compare error>> [%d]: %08x\n", i, kk);
	}
}

volatile unsigned int *d;
volatile unsigned int *s;
volatile unsigned int *pVerify;

int verify_image(int cnt, unsigned int image, unsigned int golden, int size)
{
	int i, nResult = 0;
#if 0
	volatile unsigned int *d = (volatile unsigned int *)image;
	volatile unsigned int *pVerify = (volatile unsigned int *)golden;
#else
	d = (volatile unsigned int *)image;
	pVerify = (volatile unsigned int *)golden;
#endif

	for ( i=0; i<size/4; i++) {
		if ( d[i] != pVerify[i] ) {
			printf("%4d: verify bad copy at %p [%08x] <-- %p [%08x]\n",
				   cnt, &d[i], d[i], &pVerify[i], pVerify[i]);
			nResult = -1;
		}
	}
		
	return nResult;
}

int copy_image_by_words_backwards(unsigned int dst, unsigned int src, int size)
{
	int i;
#if 0
	volatile unsigned int *d = (volatile unsigned int *)dst;
	volatile unsigned int *s = (volatile unsigned int *)src;
#else
	d = (volatile unsigned int *)dst;
	s = (volatile unsigned int *)src;
#endif

	for ( i=size/4; i>=0; i-- ) {
		d[i] = s[i];
		asm("eieio");
		if ( d[i] != s[i] )
			printf("BUG: bad copy_at %p [%08x] <-- %p [%08x]\n",
				   &d[i], d[i], &s[i], s[i]);
	}

	printf("Copied %d words\n", i);
	return 0;
}

int copy_image_using_memcpy(unsigned int dst, unsigned int src, int size)
{
	memcpy((void *)dst, (void *)src, size);

	return 0;
}

int copy_image_by_words(unsigned int dst, unsigned int src, int size)
{
	int i;
#if 0
	volatile unsigned int *d = (volatile unsigned int *)dst;
	volatile unsigned int *s = (volatile unsigned int *)src;
#else
	d = (volatile unsigned int *)dst;
	s = (volatile unsigned int *)src;
#endif

	for ( i=0; i<size/4; i++ ) {
		d[i] = s[i];
		asm("eieio");
		if ( d[i] != s[i] )
			printf("BUG: bad copy_at %p [%08x] <-- %p [%08x]\n",
				   &d[i], d[i], &s[i], s[i]);
	}

	printf("Copied %d words\n", i);
	return 0;
}

int checksum (unsigned long addr, long len)
{
	/*
	 * Simple 32-bit sum of all 32-bit words in image
	 * Should return zero if image is valid
	 */
	volatile int *pImage = (volatile int *)addr;
	int sum = 0;

	while (len ) {
		sum += *pImage++;
		len -= sizeof(unsigned long);
	}
	return(sum);
}

/*
 * Look for image signature and return start or -1 if not found
 */
unsigned long image_scan (unsigned long start) {
	volatile unsigned int *pCksum;
	struct startup_header startup_hdr;
	int cksum;
	 
	//
	// We assume that the images will all start on a 4 byte boundary
	//
	memcpy (&startup_hdr, (void *)start, sizeof(startup_hdr));
	if (startup_hdr.signature != STARTUP_HDR_SIGNATURE)
		return -1;
	printf("Found valid QNX image\n");
	
	pCksum = (volatile unsigned int *)(start + startup_hdr.startup_size - 4);
	printf("startup_hdr.signature       = %08lx\n", startup_hdr.signature);
	printf("startup_hdr.header_size     = %08x\n", startup_hdr.header_size); 
	printf("startup_hdr.startup_size    = %08lx\n", startup_hdr.startup_size);
	printf("startup_hdr.startup_vaddr   = %08lx\n", startup_hdr.startup_vaddr);
	printf("   startup checksum            = %08x\n", *pCksum);
	printf("startup_hdr.imagefs_size    = %08lx\n", startup_hdr.imagefs_size); 

	/*
	 * Display the image checksum 
	 */
	pCksum = (volatile unsigned int *)
		(start + startup_hdr.stored_size - 20);	/* Don't ask me why the 20 */
	printf("   image checksum              = %08x\n", *pCksum);
	
	memcpy (&startup_hdr, (void *)start, sizeof(startup_hdr));

	/*
	 * First perform some sanity checks
	 */
	if ( startup_hdr.startup_size > (1024 * 1024 * 1) ) {
		printf("Bad startup_size in QNX image\n");
		return -1;
	}

	/*
	 * There are two checksums, one protecting the startup image
	 * and one for the primary image.
	 */

	if ( (cksum = checksum (start, startup_hdr.startup_size)) != 0){
		printf("QNX startup checksum failed\n");
		return -1;
	}

	if ( (cksum = checksum (start+startup_hdr.startup_size, 
				  startup_hdr.stored_size-startup_hdr.startup_size)) != 0) {
		printf("QNX image checksum failed (%08x)\n", cksum);
		return -1;
	}

	/*
	 * OK, the image seems to be a valid one.  Find the entry point
	 * and return it
	 */

	return(startup_hdr.startup_vaddr);
}

void do_image_memtest(unsigned long load_addr, int stored_size)
{
	int i;
	for ( i=0; i<64; i++ ) {
		memory_fill_words(load_addr, 0, stored_size);
		verify_fill_words(load_addr, 0, stored_size);
		//		copy_image_by_words(load_addr, GOLDEN_IMAGE, stored_size);
		//		copy_image_by_words_backwards(load_addr, GOLDEN_IMAGE, stored_size);
		copy_image_using_memcpy(load_addr, GOLDEN_IMAGE, stored_size);

		verify_image(i, load_addr, GOLDEN_IMAGE, stored_size);
		memory_fill_words(load_addr, 0xffffffff, stored_size);
		verify_fill_words(load_addr, 0xffffffff, stored_size);
		//		copy_image_by_words(load_addr, GOLDEN_IMAGE, stored_size);
		//		copy_image_by_words_backwards(load_addr, GOLDEN_IMAGE, stored_size);
		copy_image_using_memcpy(load_addr, GOLDEN_IMAGE, stored_size);
		verify_image(i, load_addr, GOLDEN_IMAGE, stored_size);

		/*
		 * Try a couple more patterns
		 */
		memory_fill_words(load_addr, 0xe7031b83, stored_size);
		verify_fill_words(load_addr, 0xe7031b83, stored_size);
		memory_fill_words(load_addr, 0x7f63db78, stored_size);
		verify_fill_words(load_addr, 0x7f63db78, stored_size);

		if ( tstc() ) {
			getc();
			break;
		}
		putc('.');
	}
	printf("\n");
}

int do_bootqnx (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	unsigned long load_addr = simple_strtol(argv[1], NULL, 16);
	struct startup_header *qnx_hdr;
	unsigned long stored_size, start_addr;
	void	(*qnx_kernel)(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[]);

	/*
	 * The QNX image expects to be loaded without a PPCBOOT header
	 * So let's copy it over the header
	 */

	qnx_hdr = (struct startup_header *) (load_addr + sizeof(image_header_t));
	stored_size = qnx_hdr->stored_size;
	printf("Image stored size is %ld\n", stored_size);

	/*
	 * Copy the image, note that it's an overlap copy, so be careful
	 */
	printf("load_addr = %08lx, qnx_hdr is at %p\n", load_addr, qnx_hdr);

	stored_size = 156600;

	start_addr = image_scan(load_addr);
	if ( start_addr == -1 ) 
		return start_addr;

	qnx_kernel = (void (*)(cmd_tbl_t *, int, int, char **))start_addr;

	printf("Found QNX image: transferring control to %p\n", qnx_kernel);
	(*qnx_kernel) (cmdtp, flag, argc, argv);
	
	return 0;
}
